// File: FW_ControlAmp.ino
// Target: Arduino Mega 2560 with Ethernet Shield
// Latest change: 2019-01-12 
// The Control amplifier talks over local network, UDP with a GUI running on a PC/Laptop for
// configuration issues like setting the channel sensitivities.
// The Control Amplifier talks via serial RS422 with the FLT boards in the loudspeakers for
// adjusting the levels of Bass and Midrange, requesting status and supply voltage measurements.
// In normal use the GUI is not required and communication with the loudspeakers is silent. 
// The FLT boards in the loudspeakers spontaneously send messages in case of an error condition.

// For port / pin use see the Messages document

#define VERSIONSTRING "Control Amplifier v 20 2019-01-14"
#define VERSIONNUMBER 20

#include <SPI.h> 
#include <Ethernet.h>
#include <EthernetUdp.h>
#include "diagnostics.h"

EthernetUDP Udp;
byte mac[] = { 0x90, 0xA2, 0xDA, 0x0F, 0x6A, 0x91 }; 
IPAddress LocalIP(192, 168, 2, 25); 
unsigned int Port = 8000;  // Listen and Talk to this port

#define MessageToGUIsize 30
byte MessageToGUI[MessageToGUIsize];
#define MessageFromGUIsize 16
byte MessageFromGUI[MessageFromGUIsize];
#define MessageToSpeakerSize 3
byte MessageToSpeaker[MessageToSpeakerSize];
#define MessageFromSpeakerSize 13
byte MessageFromSpeakerLeft[MessageFromSpeakerSize];
byte MessageFromSpeakerRight[MessageFromSpeakerSize];

// Relatieve gain settings for the 8 input channels
// value 192 is 0dB gain.
byte Corrections[9]; // 0 = soft Mute
byte ChEffect, ChCenter, ChExtra;
byte Bas, Mid, StatusL, StatusR; // for cross over, Power amplifiers 
byte adSpkrVOL, adPhnVOL; // from A/D converters
byte SpkrVOL, PhnVOL; // from potmeters, after averaging
byte Command, CommandLeft, CommandRight; // from GUI
byte Channel; 
byte IRcode;
byte CounterMainsOK;
int  CounterPGALed;
int  CounterErrorLedTimeOut;
bool TalkToSpeakers, MainsOK;

#define ERRLedLeftON   PORTL &= ~0x40
#define ERRLedLeftOFF  PORTL |=  0x40
#define ERRLedRightON  PORTL &= ~0x80
#define ERRLedRightOFF PORTL |=  0x80
#define PGALedON       PORTD &= ~0x80
#define PGALedOFF      PORTD |=  0x80
#define MotorPotiCCW   PORTL |=  0x15 // PL0+2+4 high MotorPoti ccw
#define MotorPotiCW    PORTL |=  0x2A // PL1+3+5 high MotorPoti cw
#define MotorPotiOFF   PORTL &=  0xC0
#define SelectorSwitchLong  (PIND & 0x01)
#define SelectorSwitchShort (PIND & 0x02)
#define SpeakerSwitchON ((PINJ & 0x02) == 0)
#define SpeakerPowerON  PORTK &= ~0x40;
#define SpeakerPowerOFF PORTK |= 0x40

//************************************************************************************************
ISR(TIMER1_COMPA_vect) // 1 kHz interrupt, takes ca. 10 usec.
{
  static int CounterLedTest = 2000;
  static bool Blink;
  static int CounterIRTimeOut, CounterBlinkTime;
  
DIAGpin11H // diagnose interrupt rate and duration

// Manage Leds
  if(CounterBlinkTime > 0) CounterBlinkTime--; else {Blink = !Blink; CounterBlinkTime = 200;}
  
  if(CounterLedTest > 0)
  {
    ERRLedLeftON;
    ERRLedRightON;
    PGALedON;
    CounterLedTest--;
    return;
  }

// Manage PGA Led
// PGA led is on while potmeters are changing
// PGA led blinks in Mute condition
  static bool Mute;
  if(CounterPGALed > 0)
  {
    PGALedON; 
    CounterPGALed--; 
  } 
  else 
    if(Mute && Blink) PGALedON; // PGAled blinks in mute mode
      else PGALedOFF;

// Manage ERRor leds    
  if(StatusL > 0) ERRLedLeftON; else ERRLedLeftOFF;
  if(StatusR > 0) ERRLedRightON; else ERRLedRightOFF;
  // CounterErrorLedTimeOut is set to 100 by a spontaneous message
  if(CounterErrorLedTimeOut > 0) CounterErrorLedTimeOut--;
  if(CounterErrorLedTimeOut == 1) // only once 
  {
    ERRLedLeftOFF; 
    ERRLedRightOFF;
    StatusL = StatusR = 0; 
  }
  
// Manage IR 
  static bool MuteDone;
  if((PINK & 0x80) == 0x80) // VT high = IR active
  {
    IRcode = (PINC >> 4) & 0x0F;
    if((IRcode <= 8) && (IRcode > 0)) 
      {
        Channel = IRcode;
        SetChannel();
      }
    if(IRcode == 9)  { CounterIRTimeOut = 200; MotorPotiCCW; }    
    if(IRcode == 10) { CounterIRTimeOut = 200; MotorPotiCW; }
    if(IRcode == 0)  
      if(!MuteDone)
      { 
        Mute = !Mute; 
        MuteDone = true; // prevent multiple toggles
      } 
  }
  else // VT low = IR not active
  {
    if(CounterIRTimeOut > 0) // wait some time
      CounterIRTimeOut--;
    else 
    {
      MotorPotiOFF;
      MuteDone = false;
    }
  }  

// Manage Channel selection switch 
// The process starts when the long pulse goes low
// When in the next pass we see the short pulse low it is CW, else CCW  
// The process ends when the long pulse goes high
static bool Started, Done;
  if(Started && !Done)  
  {
    if(SelectorSwitchLong == 0) Channel++; else Channel--;
    if(Channel == 0) Channel = 1; 
    if(Channel > 8) Channel = 8;
    Done = true;
    SetChannel();
  }
  if(SelectorSwitchShort == 0 && !Started) Started = true; // Long pulse goest low
  if(Done)
    if(SelectorSwitchShort != 0) // wait for long pulse high
    {
      Started = false;
      Done = false;
    } 
  
// Manage Potmeters
  #define Margin 2 
  static bool MeasSpkr; // Are we measuring the Speaker potmeter or the Phone potmeter
  static unsigned int avrSpkrVOL, avrPhnVOL; // for averaging
  static byte PrevSpkrVOL, PrevPhnVOL; // prevous values
  int iadSpkrVOL, iadPhnVOL;
  if(MeasSpkr)
  {
    iadSpkrVOL = analogRead(A8);
    adSpkrVOL = iadSpkrVOL >> 2;
    if(Mute) adSpkrVOL = 0;
    // average with leaking bucket algorithm
    avrSpkrVOL -= (avrSpkrVOL >> 4);  // subtract 1/16 of highbyte 
    avrSpkrVOL += (int(adSpkrVOL) << 4); // add 1/16 of new value
    SpkrVOL = avrSpkrVOL >> 8; // take highbyte
    if(abs(int(SpkrVOL) - int(PrevSpkrVOL)) > Margin) // typecast b.o. negative
    {
      CounterPGALed = 200;
      UpdatePGAs();
      PrevSpkrVOL = SpkrVOL; 
    }
  }
  else
  {
    iadPhnVOL = analogRead(A9);
    adPhnVOL = iadPhnVOL >> 2;
    avrPhnVOL -= (avrPhnVOL >> 4); 
    avrPhnVOL += (int(adPhnVOL) << 4);
    PhnVOL = avrPhnVOL >> 8;
    if(abs(int(PhnVOL) - int(PrevPhnVOL)) > Margin) // typecast b.o. negative 
    {
      CounterPGALed = 200;
      UpdatePGAs();
      PrevPhnVOL = PhnVOL;
    }
  }
  MeasSpkr = !MeasSpkr;

// Manage MainsOK
// CounterMainsOK is set to 15 at each mains zero crossing
  if(CounterMainsOK > 0) 
  {
    CounterMainsOK--;
    MainsOK = true;
  }
  else 
  {
    MainsOK = false;
    WriteEEPROM(); 
  }
   
// Manage speaker power switch
  if(SpeakerSwitchON && MainsOK) SpeakerPowerON else SpeakerPowerOFF; 
  
DIAGpin11L // diagnose timer interrupt handling
} 

//************************************************************************************************
void EvalMessageFromGUI(void)
{
  Command = MessageFromGUI[0];
  memcpy (&Corrections[1], &MessageFromGUI[1], 8);
  ChEffect = MessageFromGUI[9];
  ChCenter = MessageFromGUI[10];
  ChExtra = MessageFromGUI[11];
  SetChannel();
  CommandLeft = MessageFromGUI[12];
  CommandRight = MessageFromGUI[13];
  Bas = MessageFromGUI[14];
  Mid = MessageFromGUI[15];
  if((Command & 0x01)== 0x01) TalkToSpeakers = true;
}

//************************************************************************************************
void PrepareMessageToGUI(void)
{
  MessageToGUI[0] = Channel;
  MessageToGUI[1] = IRcode;
  MessageToGUI[2] = adSpkrVOL;
  MessageToGUI[3] = adPhnVOL;
  memcpy(&MessageToGUI[4],  &MessageFromSpeakerLeft,  MessageFromSpeakerSize); 
  memcpy(&MessageToGUI[17], &MessageFromSpeakerRight, MessageFromSpeakerSize);
}

//************************************************************************************************
void SetChannel(void)
{
  PORTC &= 0xF0; PORTC |= Channel; // to Nixie as Nibble
  if (Channel > 0) PORTA = ~(1 << (Channel - 1)); // to Selection PCB. Low = select
  UpdatePGAs();
}

//************************************************************************************************
byte Map(byte UpperLimit, byte Entry, byte LowerLimit)
{
  byte bmap = LowerLimit + Entry * (UpperLimit - LowerLimit) / 255; 
  if (Entry == 0) bmap = 0; // force the soft mute condition 
  return bmap;
}

//************************************************************************************************
void SendByteToPGA(byte V) // send a single byte
{
  for(byte i=0; i<8; i++)
  {
    if((V & 0x80) == 0x80) PORTG |= 0x02; else PORTG &= ~0x02;  // Databit
    PORTG |= 0x01; // Clock high
    V = V << 1;  
    PORTG &= ~0x01; // Clock low
  }  
}

//************************************************************************************************
void UpdatePGAs(void)
{
  byte Correction, SpkrVol, PhnVol, Effect, Center, Extra;
  // Map Volume poti to range 0 .. -50 dB
  SpkrVol = Map(192, SpkrVOL, 92);
  // Map TEL poti to range +20 .. -30 dB
  PhnVol = Map(232, PhnVol, 132);
  if (adPhnVOL < 3) PhnVol = 0; // soft mute
  Effect = SpkrVol + (ChEffect - 192);
  Center = SpkrVol + (ChCenter - 192);
  Extra = SpkrVol + (ChExtra - 192);
  if((adSpkrVOL < 3) || (Channel == 0)) SpkrVol = Effect = Center = Extra = 0; // soft mute
  Correction = Corrections[Channel];
  // PGA's are daisychained so the order matters!
  PORTG &= ~0x04; // CS low 
  SendByteToPGA(Extra);  // Extra channel 
  SendByteToPGA(Center); // Center loudspeaker
  SendByteToPGA(Effect);     SendByteToPGA(Effect);  // effect channels
  SendByteToPGA(PhnVol);     SendByteToPGA(PhnVol);  // volume headphone
  SendByteToPGA(SpkrVol);    SendByteToPGA(SpkrVol); // volume loudspeakers
  SendByteToPGA(Correction); SendByteToPGA(Correction); // input correction
  PORTG |= 0x04; // CS high 
}

/**************************************************************************************************/
unsigned char EEPROM_readByte(unsigned int uiAddress)
{
  // Wait for completon of previous action
  while(EECR & (1<<EEPE));
  EEAR = uiAddress;
  EECR |= (1<<EERE);
  return EEDR;
}

/**************************************************************************************************/
void EEPROM_writeByte(unsigned int uiAddress, unsigned char ucData)
{
  // Wait for completon of previous action
  while(EECR & (1<<EEPE));
  EEAR = uiAddress;
  EEDR = ucData;
  EECR |= (1<<EEMPE);
  EECR |= (1<<EEPE);
}

//********************************************************************************************************
void ReadEEPROM(void)
{
  Channel =  EEPROM_readByte(0);
  for (byte i=1; i<9; i++) Corrections[i] = EEPROM_readByte(i);
  ChEffect =  EEPROM_readByte(9); 
  ChCenter =  EEPROM_readByte(10);
  ChExtra =  EEPROM_readByte(11);
}

//********************************************************************************************************
void WriteEEPROM(void)
{
  EEPROM_writeByte(0, Channel);
  for (byte i=1; i<9; i++) EEPROM_writeByte(i, Corrections[i]);
  EEPROM_writeByte(9, ChEffect); 
  EEPROM_writeByte(10, ChCenter);
  EEPROM_writeByte(11, ChExtra);
}

//********************************************************************************************************
ISR(PCINT1_vect)
// Interrupt from mains comes on each edge, so 10 ms 
// For Power Down detection >> store data in EEPROM
{
  CounterMainsOK = 15; 
  //DIAGpin10T // diagnose
  // Is decremented in ISR(TIMER1_COMPA_vect) at 1 kHz rate
}
 
//********************************************************************************************************
void setup()
{ 
  Serial.begin(19200);
  Serial.println(VERSIONSTRING);
  Serial1.begin(19200); // Spkr right
  Serial2.begin(19200); // Spkr left
    
  // PCINT1 on port J0 (also RXD3, not used) is PCINT9 for 100 Hz mains detection
  PCICR |= 0x02; // PCIE1 for PCINT 8..15
  PCMSK1 |= 0x02; // PCINT9 = PJ0
  
  // I/O ports
  DDRA = 0xFF;   // Outputs to Selector PCB
  DDRC = 0x0F;   // Lower nibble is Nixie Out
  PORTJ |= 0x02; // Pull-up for Speakerpower switch
  DDRD = 0x80;   // LED PGA, Mute 
  DDRK |= 0x40;  // Speaker Power
  DDRL = 0xFF;   // Outputs forr MotorPoti and Error Leds
  DDRG = 0x07;   // For PGA's
 
  // configure 16 bit Timer1 for Interrupt ISR(TIMER1_COMPA_vect)
  TCCR1A = 0;     // needs to be done
  TCCR1B = 0x0A;  // select internal clock, prescaler 8, WaveformMode 4
  OCR1A = 1990;   // 1 msec interval = 1000 Hz
  TIMSK1 = 0x02;  // interrupt on OCRA match  
 
  // Initialisation
  ReadEEPROM();
  Corrections[0] = 0; // Channel[0] = Mute condition
  SetChannel();
  Serial.print("Channel corrections ");
  for (byte i=1; i<9; i++) {Serial.print(Corrections[i]); Serial.print(" ");}
  Serial.println();
  char str[50];
  sprintf(str, "Channel: %3d  Eff: %3d Centr: %3d  Xtra: %3d", Channel, ChEffect, ChCenter, ChExtra);
  Serial.println(str);
  
  Ethernet.begin(mac, LocalIP);
  byte dummy = Udp.begin(Port); 
  
  UseDIAGpin10
  UseDIAGpin11
  UseDIAGpin12
  UseDIAGpin13
}

//********************************************************************************************************
void loop()
{  
// Manage UDP communication with GUI
  int packetSize = Udp.parsePacket();
  if(packetSize == MessageFromGUIsize)
  {  
    Udp.read(&MessageFromGUI[0], MessageFromGUIsize);
    EvalMessageFromGUI();
    PrepareMessageToGUI();
    Udp.beginPacket(Udp.remoteIP(), Port);
    Udp.write(&MessageToGUI[0], MessageToGUIsize);
    Udp.endPacket();
    MessageFromSpeakerLeft[0] = 0; // clear indication fresh message
    MessageFromSpeakerRight[0] = 0; // clear indication fresh message
  }
  
// Manage communication with loudspeakers
// The FLT's only talk when they were spoken to, or when an error condition exists
  
  if(Serial1.available()) // from Right speaker
  {
    delay(10); // wait until the message is complete 
    Serial1.readBytes(MessageFromSpeakerRight, MessageFromSpeakerSize);
    StatusR = MessageFromSpeakerRight[0];
    if(!TalkToSpeakers && ((StatusR & 0x40) == 0x40)) CounterErrorLedTimeOut = 100; // To switch Led Off after some time
   }
   
  if(Serial2.available()) // from left speaker
  {
    delay(10); // wait until the message is complete
    Serial2.readBytes(MessageFromSpeakerLeft, MessageFromSpeakerSize);
    StatusL = MessageFromSpeakerLeft[0];
    if(!TalkToSpeakers && ((StatusL & 0x40) == 0x40)) CounterErrorLedTimeOut = 100; // To switch Led Off after some time
  }
  
  if(TalkToSpeakers)
   {
    Serial1.write(CommandRight);
    Serial1.write(Bas);  
    Serial1.write(Mid); 
    Serial2.write(CommandLeft);
    Serial2.write(Bas);
    Serial2.write(Mid); 
    TalkToSpeakers = false;
  }
}
